//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Descriptor/PolyItem.h
//! @brief     Defines and implements template class PolyItem.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2021
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_GUI_MODEL_DESCRIPTOR_POLYITEM_H
#define BORNAGAIN_GUI_MODEL_DESCRIPTOR_POLYITEM_H

#include "GUI/Model/Util/UtilXML.h"

//! Holds a polymorphous item. Possible types of the item are specified by a Catalog.

template <typename Catalog> class PolyItem {
public:
    using BaseItem = typename Catalog::BaseItem;

    void simpleInit(const QString& label, const QString& tooltip,
                    typename Catalog::Type currentType);

    BaseItem* operator->() const { return m_item.get(); }
    BaseItem* certainItem() const { return m_item.get(); }

    void setCertainItem(BaseItem* t) { m_item.reset(t); }

    template <typename S> S* createCertainItem();

    //! Serializes the catalog index of the currently selected type and calls
    //! main serialization method of the selected class.
    void writeTo(QXmlStreamWriter* w) const { XML::writeItemTo<Catalog>(m_item.get(), w); }

    template <typename... Args> BaseItem* readItemFrom(QXmlStreamReader* r, Args... args);

    //! Deserializes the catalog index of the currently selected type, creates a new
    //! object of this type and calls main deserialization method of the selected class.
    template <typename... Args> void readFrom(QXmlStreamReader* r, Args... args)
    {
        m_item.reset(readItemFrom(r, args...));
    }

    QString label() const { return m_label; }
    QString tooltip() const { return m_tooltip; }
    QStringList options() const { return m_options; }

    void setCurrentIndex(int newIndex) { currentIndexSetter(newIndex); }
    int currentIndex() const { return m_types.indexOf(Catalog::type(m_item.get())); }

private:
    void initFieldsAndSetter(const QString& label, const QString& tooltip);

    //! Reimplementable function to set currently selected option.
    std::function<void(int)> currentIndexSetter = nullptr;

    std::unique_ptr<BaseItem> m_item; //!< Current selection

    QString m_label;       //!< A label text (short, no trailing colon)
    QString m_tooltip;     //!< Tooltip text
    QStringList m_options; //!< List of options, usually presented as combo entries
    QVector<typename Catalog::Type> m_types = Catalog::types();
};


//! Initialize by means of a catalog class and optional creation arguments.
//!
//! The current selection will be initialized with the first type in the catalog types. The optional
//! arguments are the arguments which may be necessary for the creation method in the catalog.
template <typename C>
void PolyItem<C>::simpleInit(const QString& label, const QString& tooltip,
                             typename C::Type currentType)
{
    initFieldsAndSetter(label, tooltip);
    int index = C::types().indexOf(currentType);
    setCurrentIndex(index);
}

//! Directly set the new item.
template <typename C> template <typename S> S* PolyItem<C>::createCertainItem()
{
    auto* t = new S;
    m_item.reset(t);
    return t;
}

template <typename C>
template <typename... Args>
C::BaseItem* PolyItem<C>::readItemFrom(QXmlStreamReader* r, Args... args)
{
    const uint typeIndex = XML::readUInt(r, XML::Attrib::type);
    const QString kind = XML::readString(r, XML::Attrib::kind);
    const auto type = static_cast<typename C::Type>(typeIndex);
    BaseItem* t = C::create(type, args...);
    if (t)
        t->readFrom(r);
    return t;
}

template <typename C>
void PolyItem<C>::initFieldsAndSetter(const QString& label, const QString& tooltip)
{
    m_label = label;
    m_tooltip = tooltip;
    m_options.clear();
    for (const auto type : m_types)
        m_options << C::uiInfo(type).menuEntry;

    currentIndexSetter = [this](int current) {
        BaseItem* t = C::create(m_types[current]);
        m_item.reset(t);
    };
}

#endif // BORNAGAIN_GUI_MODEL_DESCRIPTOR_POLYITEM_H
